<?php
namespace Util;

use Moment\Moment;
use Cron\CronExpression;
use Zeedhi\Framework\DependencyInjection\InstanceManager;

class Period {

    const PREFIX_THIS       = 'this';
    const PREFIX_NEXT       = 'next';
    const PREFIX_LAST       = 'last';
    public static $PREFIXES = array(self::PREFIX_THIS, self::PREFIX_NEXT, self::PREFIX_LAST);

    const SUFFIX_INCLUSIVE  = 'inclusive';
    public static $SUFFIXES = array(self::SUFFIX_INCLUSIVE);

    const UNIT_DAY       = "day";
    const UNIT_WEEK      = "week";
    const UNIT_MONTH     = "month";
    const UNIT_YEAR      = "year";
    public static $UNITS = array(self::UNIT_DAY, self::UNIT_WEEK, self::UNIT_MONTH, self::UNIT_YEAR);

    const PREPOSITION_IN    = 'in';
    const PREPOSITION_UNTIL = 'until';
    public static $PREPOSITIONS = array(self::PREPOSITION_IN, self::PREPOSITION_UNTIL);

    const DATE_FORMAT = 'd/m/Y';

    public static $EXCEPTION_UNITS = array(
        'days'   => self::UNIT_DAY,
        'weeks'  => self::UNIT_WEEK,
        'months' => self::UNIT_MONTH,
        'years'  => self::UNIT_YEAR
    );

    public static $ONE_WORD_EXPRESSIONS = array(
        'today' 	=> 'this-day',
        'yesterday'	=> 'last-day',
        'tomorrow'	=> 'next-day'
    );

    public static function decode($period,$dateBase='') {
        if(empty($dateBase)){
            $start = new Moment('now', "America/Sao_Paulo");
            $end = new Moment('now', "America/Sao_Paulo");
        }else{
            $start = new Moment($dateBase, "America/Sao_Paulo");
            $end = new Moment($dateBase, "America/Sao_Paulo");
        }
        try {
            if($period) {
                if(preg_match('/^[0-9a-f]{24}$/i', $period) === 1) {
                    /** @var $periodsService \Service\PeriodsService */
                    $periodsService = InstanceManager::getInstance()->getService('\Service\PeriodsService');
                    $periodMongo = $periodsService->findById($period);
                    if(empty($periodMongo)){
                        $period = 'today';
                    }
                }
                if (!empty($periodMongo)) {
                    list($start, $end) = self::getPeriodMathExpression($periodMongo,$dateBase);
                } else {
                    $periodInformations = self::getPeriodParams($period);
                    foreach ($periodInformations as $params) {
                        $period = self::buildInterval($params, $start, $end,$dateBase);
                        $start = $period['start'];
                        $end = $period['end'];
                    }
                }
            }
            $interval = array($start, $end);
            return $interval;
        } catch (\Exception $e) {
            var_dump($e->getMessage());
        }
    }

    protected static function getPeriodParams($period){
        $periodInformations = array();
        $period = self::conditionPeriod($period);
        $expressions = explode('|', $period);
        foreach ($expressions as $expression) {
            if(strlen($expression)>0) {
                $param = self::decodeExpression($expression);
                array_push($periodInformations, $param);
            }
        }

        return $periodInformations;
    }

    protected static function decodeExpression($expression){
        $splitedExpression = array();
        $expressionTerms = explode('-', $expression);
        foreach ($expressionTerms as $terms) {
            if (array_key_exists($terms, self::$ONE_WORD_EXPRESSIONS)) {
                $terms = self::$ONE_WORD_EXPRESSIONS[$terms];
            }
            $terms = explode('-', $terms);

            foreach ($terms as $term) {
                array_push($splitedExpression, $term);
            }
        }

        $decodedExpression = self::recognizeExpressionTerms($splitedExpression);
        return $decodedExpression;
    }

    protected static function recognizeExpressionTerms($splitedExpression){
        $decodedExpression = array();
        foreach ($splitedExpression as $term) {
            if (array_key_exists($term, self::$EXCEPTION_UNITS)) {
                $term = self::$EXCEPTION_UNITS[$term];
            }
            if(in_array($term, self::$PREPOSITIONS)){
                $decodedExpression['preposition'] = $term;
            } else if(in_array($term, self::$PREFIXES)){
                $decodedExpression['prefix'] = $term;
            } else if(in_array($term, self::$UNITS)){
                $decodedExpression['unit'] = $term;
            } else if(in_array($term, self::$SUFFIXES)){
                $decodedExpression['suffix'] = $term;
            } else if(!is_nan($term)){
                $decodedExpression['amount'] = $term;
            }
        }

        return $decodedExpression;
    }

    protected static function buildInterval($params, $start, $end,$dateBase =''){
        $preposition = isset($params['preposition']) ? $params['preposition'] : null;
        $prefix      = isset($params['prefix'])      ? $params['prefix']      : null;
        $unit        = isset($params['unit'])        ? $params['unit']        : null;
        $suffix      = isset($params['suffix'])      ? $params['suffix']      : null;
        $amount      = isset($params['amount'])      ? $params['amount']      : 1;

        switch ($prefix) {
            case self::PREFIX_LAST:
                return self::buildIntervalToLastPrefix($preposition, $amount, $unit, $suffix, $start, $end, $dateBase);
            case self::PREFIX_NEXT:
                return self::buildIntervalToNextPrefix($preposition, $amount, $unit, $suffix, $start, $end, $dateBase);
            case self::PREFIX_THIS:
                return self::buildIntervalToThisPrefix($preposition, $unit, $suffix, $start, $end, $dateBase);
            case null:
                return self::buildIntervalToNoPrefix($preposition, $amount, $unit, $suffix, $start, $end);
        }
    }

    protected static function buildIntervalsToSpecialCase(&$start, &$end, $specialCase, $preposition){
        if($specialCase && $preposition !== self::PREPOSITION_UNTIL){
			self::subtractDate($start, self::UNIT_MONTH, 1);
			self::subtractDate($end, self::UNIT_MONTH, 1);
		}
    }

    protected static function buildIntervalToLastPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd, $dateBase = ''){
        $endAmount = $suffix === self::SUFFIX_INCLUSIVE ? 0 : 1;
        $startAmount = $suffix === self::SUFFIX_INCLUSIVE ? $amount - 1 : $amount;

        if(empty($dateBase)){
            $start = new Moment('now', "America/Sao_Paulo");
            $end = new Moment('now', "America/Sao_Paulo");
        }else{
            $start = new Moment($dateBase, "America/Sao_Paulo");
            $end = new Moment($dateBase, "America/Sao_Paulo");
        }

        self::subtractDate($start, $unit, $startAmount);
        self::subtractDate($end, $unit, $endAmount);

        if($preposition === self::PREPOSITION_IN) {
            $periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
            $periodEnd = self::getMomentWhenInClause($periodEnd, $unit, $end);
        } else {
            $periodEnd = $end->endOf($unit);
            if(!$preposition){
                $periodStart = $start->startOf($unit);
            }
        }

        if($periodStart > $periodEnd){
            $periodEnd = self::copyDate($periodStart);
        }

        $interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function copyDate(Moment $data){
        return new Moment($data->format('Y-m-d'));
    }

    protected static function buildIntervalToNextPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd, $dateBase= ''){
        $startAmount = $suffix === self::SUFFIX_INCLUSIVE ? 0 : 1;
		$endAmount   = $suffix === self::SUFFIX_INCLUSIVE ? $amount - 1 : $amount;

        if(empty($dateBase)){
            $start = new Moment('now', "America/Sao_Paulo");
            $end = new Moment('now', "America/Sao_Paulo");
        }else{
            $start = new Moment($dateBase, "America/Sao_Paulo");
            $end = new Moment($dateBase, "America/Sao_Paulo");
        }

		self::addDate($start, $unit, $startAmount);
		self::addDate($end, $unit, $endAmount);

		if (!$preposition) {
			$periodStart = $start->startOf($unit);
			$periodEnd   = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_UNTIL){
		    $periodEnd = $end->startOf($unit);
		}
		else if ($preposition === self::PREPOSITION_IN) {
			$periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
			$periodEnd = self::getMomentWhenInClause($periodEnd, $unit, $end);
		}

		$interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function buildIntervalToThisPrefix($preposition, $unit, $suffix, $periodStart, $periodEnd, $dateBase =''){
        if(empty($dateBase)){
            $start = new Moment('now', "America/Sao_Paulo");
            $end = new Moment('now', "America/Sao_Paulo");
        }else{
            $start = new Moment($dateBase, "America/Sao_Paulo");
            $end = new Moment($dateBase, "America/Sao_Paulo");
        }


        if (!$preposition) {
			$periodStart = $start->startOf($unit);
			$periodEnd = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_UNTIL) {
			if($suffix !== self::SUFFIX_INCLUSIVE){
				self::subtractDate($end, $unit, 1);
			}
			$periodEnd = $end->endOf($unit);
		}
		else if ($preposition === self::PREPOSITION_IN) {
			$periodStart = self::getMomentWhenInClause($periodStart, $unit, $start);
			$periodEnd   = self::getMomentWhenInClause($periodEnd, $unit, $end);
		}

		$interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function buildIntervalToNoPrefix($preposition, $amount, $unit, $suffix, $periodStart, $periodEnd){
        if(!$preposition){
            $periodStart = self::setDateParam($unit, $amount, $periodStart);
        } else if($preposition === self::PREPOSITION_UNTIL){
            $periodEnd = self::setDateParam($unit, $amount, $periodEnd);
        }

        $interval = self::treatInterval($periodStart, $periodEnd, $unit);
        return $interval;
    }

    protected static function setDateParam($unit, $amount, Moment $moment) {
		switch($unit){
			case self::UNIT_DAY:
				return $moment->setDay($amount);
			case self::UNIT_WEEK:
				return $moment->setWeek($amount);
			case self::UNIT_MONTH:
                $month = $amount+1;
                $date = \DateTime::createFromFormat('d/m/Y H:i:s',"01/$month/".$moment->getYear()." 00:00:00");
                $date->modify('last day of this month');
                $lastday = $date->format('d');
                if($moment->getDay() > $lastday){
                    $moment->setDay($lastday);
                }

				return $moment->setMonth($month);
			case self::UNIT_YEAR:
				return $moment->setYear($amount);
		}
	}

    protected static function getMomentWhenInClause($moment, $unit, $newMoment){
        if($unit === self::UNIT_DAY){
            $moment = $newMoment;
        }
        else if ($unit === self::UNIT_MONTH){
            $moment->setMonth($newMoment->getMonth());
            $moment->setYear($newMoment->getYear());
        }
        else if ($unit === self::UNIT_YEAR){
            $moment->setYear($newMoment->getYear());
        }

        return $moment;
    }

	public static function subtractDate(Moment &$date, $unit, $qtd = 1) {
        switch ($unit) {
            case self::UNIT_DAY:
                $date->subtractDays($qtd);
                break;
            case self::UNIT_WEEK:
                $date->subtractWeeks($qtd);
                break;
            case self::UNIT_MONTH:
                self::checkLastDayOfMonths($date, $qtd);
                $date->subtractMonths($qtd);
                break;
            case self::UNIT_YEAR:
                $date->subtractYears($qtd);
                break;
        }
    }

    protected static function checkLastDayOfMonths(&$date, $qtd){
        $newDate = clone $date;
        $newMonthDate = $newDate->startOf(self::UNIT_MONTH)->subtractMonths($qtd);
        $lastDayOfNewMonth = $newMonthDate->endOf(self::UNIT_MONTH)->getDay();
        $nowDay = $date->getDay();
        if($nowDay > $lastDayOfNewMonth){
            $date->setDay($lastDayOfNewMonth);
        }
    }

    public static function addDate(Moment &$date, $unit, $qtd = 1) {
        switch ($unit) {
            case self::UNIT_DAY:
                $date->addDays($qtd);
                break;
            case self::UNIT_WEEK:
                $date->addWeeks($qtd);
                break;
            case self::UNIT_MONTH:
                $date->addMonths($qtd);
                break;
            case self::UNIT_YEAR:
                $date->addYears($qtd);
                break;
        }
    }

    public static function formatInterval(array $interval) {
        return array (
            $interval[0]->format(self::DATE_FORMAT),
            $interval[1]->format(self::DATE_FORMAT)
        );
    }

    public static function convertStringToMoment($periodStart, $periodEnd, $format = self::DATE_FORMAT) {
        try {
            $dateTimeStart = \DateTime::createFromFormat($format, $periodStart);
            $dateTimeEnd = \DateTime::createFromFormat($format, $periodEnd);
            $start = self::convertDateTimeToMoment($dateTimeStart);
            $end = self::convertDateTimeToMoment($dateTimeEnd);

            $start->startOf(self::UNIT_DAY);
            $end->endOf(self::UNIT_DAY);

            return array($start, $end);
        } catch(\Exception $e) {
			var_dump($e->getMessage());
		}
    }

    public static function convertDateTimeToMoment(\DateTime $dateTime) {
        $moment = new Moment($dateTime->format('Y-m-d'));
        return $moment;
    }

    public static function decodeFormatted($period,$dateBase = '') {
        $now = new Moment('now', "America/Sao_Paulo");
        if($period == "last-month|until-last-day|in-last-month" && $now->format('DD') == '01'){
            $period = "until-last-day|in-last-month";
        }
        $period = self::decode($period,$dateBase);
        return self::formatInterval($period);
    }

    protected static function treatInterval(Moment $start, Moment $end, $unit) {
        if ($unit === self::UNIT_WEEK) {
            $start->subtractDays();
            $end->subtractDays();
        }

        return array(
            'start' => $start,
            'end' => $end
        );
    }

    public static function getNextRunDateFromExpression($expression) {
        $cron = CronExpression::factory($expression);
        return $cron->getNextRunDate();
    }

    public static function conditionPeriod($period){
        $periodArray = explode('(', $period);
        if (in_array("if ",$periodArray)) {
            $condition = str_replace('new Date().getDate()', 'date("d")', $period);
            $condition = eval($condition);
            return $condition;
        } else {
            return $period;
        }
    }

    /**
     * @param $period
     * @param string $dateBase
     * @return array
     */
    protected static function getPeriodMathExpression($period,$dateBase = ''): array
    {
        $startDateExpression = !empty($period['START_DATE_EXPRESSION'])?$period['START_DATE_EXPRESSION']:'';
        $endDateExpression = !empty($period['END_DATE_EXPRESSION'])?$period['END_DATE_EXPRESSION']:'';

        $start = \Service\PeriodMathExpression\Period::getDateMoment($startDateExpression,$dateBase);
        $end = \Service\PeriodMathExpression\Period::getDateMoment($endDateExpression,$dateBase);
        return array($start, $end);
    }
}

